# coding=utf8
#####################################################################
#    Program pro bending korelační / kovarianční  matice            #
#      programovací jazyk Python 3                                  #
#      pro běh programu nutné naistalovat knihovnu NumPy            #
#      dle metody Jorjani et al. (2003)                             #
#             autoři: Veselá Z., Vostrý L., Brzáková M., Novotná A. #
#                                                  verze 1.00, 2023 #
#####################################################################

import numpy as np

# Kontrola symetrie matic
def isSymmetric(mat,N):
    transmat = np.array(mat).transpose()
    if np.array_equal(mat, transmat):
        return True
    return False

print()
np.set_printoptions(suppress=True)
# Načítá korelační matici
try:
    m = np.loadtxt('matice', dtype=float)
except ValueError as error:
    print('CHYBA!')
    print('Opravte matici "matice":')
    print(error)
    print()
    exit()
except FileNotFoundError:
    print('CHYBA!')
    print('Soubor "matice" nenalezen.')
    print()
    exit()
if (np.trace(m) / m.shape[0] != 1) and ((np.sum(m) - np.trace(m) / (m.size - m.shape[0])) > 1):
    print('Zadaná kovarianční matice:')
    podm='cov'
else:
    print('Zadaná korelační matice:')
    podm='cor'
print(m)
print()
# Načítá matici vah
try:
    vahy = np.loadtxt('vahy', dtype=float)
except ValueError as error:
    print('CHYBA!')
    print('Opravte matici vah "vahy":')
    print(error)
    print()
    exit()
except FileNotFoundError:
    print('CHYBA!')
    print('Soubor "vahy" nenalezen.')
    print()
    exit()
print('Zadaná matice vah:')
print(vahy)
print()

# Kontrola kovarianční / korelační matice
sym = isSymmetric(m, m.shape[0])
if sym==False:
    print('CHYBA!')
    print('Opravte matici "matice".')
    print('Zadaná matice není symetrická.')
    print()
    exit()
# Kontrola matice vah
sym = isSymmetric(vahy, m.shape[0])
if sym==False:
    print('CHYBA!')
    print('Opravte matici vah "vahy".')
    print('Zadaná matice není symetrická.')
    print()
    exit()
if m.shape[0] != vahy.shape[0]:
    print('CHYBA!')
    print('Opravte dosazované matice.')
    print('Korelační / kovarianční matice a matice vah mají odlišné rozměry.')
    print()
    exit()

# Převod matice vah pokud jsou zadány počty jedinců
prum_vah = (np.sum(vahy) - np.trace(vahy)) / (m.size - m.shape[0])
if prum_vah > 1:
    print('Průměrné mimodiagonální hodnoty v matici vah: ' + str(prum_vah) + ' > 0')
    print('Pro vážení při bendingu budou váhy převedeny na reciproční hodnoty.')
    w = np.full((m.shape[0],m.shape[0]), 0, np.float32)
    for i in range (0, m.shape[0]):
        for j in range (0, m.shape[0]):
            w[i,j]=1/vahy[i,j]
    print('Váhy použité při bendingu: ')
    print(w)
elif prum_vah == 1:
    print('Průměrné mimodiagonální hodnoty v matici vah = ' + str(prum_vah))
    print('Bending proběhne bez vážení.')
    w = vahy
else:
    print('Průměrné mimodiagonální hodnoty v matici vah: ' + str(prum_vah) + ' < 0')
    print('Pro vážení při bendingu budou použity zadané hodnoty.')
    w = vahy
print()

# Bending dle Jorjani et al. (2003)
r=m
round=1
while(np.all(np.linalg.eigvals(r) > 0) != True):
    round=round+1
    b, u = np.linalg.eig(r) # eigenvalues (b) a eigenvectors (u)
    b1=[]
    bb=[]
    if podm=='cov':
        for i in range (0, m.shape[0]):
            bb.append(b[i])
            if bb[i]<0.0001:
                bb[i]=0.0001
        # Výpočet nové kovarianční matice
        r2=r - (r - (u @ np.diag(bb) @ np.transpose(u))) * w
    if podm=='cor':
        for i in range (0, m.shape[0]):
            b1.append(b[i])
            bb.append(0)
            if b1[i]<0.0001:
                b1[i]=0.0002
        for i in range (0, m.shape[0]):
            bb[i]=b1[i]*(np.sum(b)/np.sum(b1))
        # Výpočet nové korelační matice
        r1=r - (r - (u @ np.diag(bb) @ np.transpose(u))) * w
        # Úprava korelační matice
        r2 = np.full((m.shape[0],m.shape[0]), 0, np.float32)
        for i in range (0, m.shape[0]):
            for j in range (0, m.shape[0]):
                r2[i,j]=r1[i,j]/np.sqrt(r1[i,i]*r1[j,j])
    r=r2

print('Počet iterací: ' + str(round))
print()

if podm=='cov':
    np.savetxt('cov-bend', r, fmt='%.10f')
    print('Pozitivně definitní kovarianční matice:')
    print(r)
    print()
    dif=m-r
    print('Odchylky kovarianční matice po bendingu:')
    print(dif)
    print()
    # Výpočet korelační matice z matice kovarianční
    diag = np.diag(1/np.sqrt(np.diag(r)))
    r = diag @ r @ diag

np.savetxt('cor-bend', r, fmt='%.10f')
print('Pozitivně definitní korelační matice:')
print(r)
print()

if podm=='cov':
    cov = m
    diag = np.diag(1/np.sqrt(np.diag(cov)))
    m = diag @ cov @ diag
dif=m-r
print('Odchylky korelační matice po bendingu:')
print(dif)
print()
